/*
** xmsgsaver ver 1
**
** saver.c -- main routine
**
** Sep 5,1999 by T.Kudoh(CDR/TK)
**
** Sep 5,1999 -- Ver 1.2
** Sep 7,1999 -- Ver 1.3
*/

#include<stdio.h>
#include<X11/Xatom.h>
#include<X11/Intrinsic.h>
#include<X11/StringDefs.h>
#include<X11/extensions/scrnsaver.h>
#include"font.h"

char *Myname;

typedef struct{
  Pixel foreground;
  Pixel background;
  String message;
  String vfontcap;
  String vectorFont;
  int fontWidth;
  int fontHeight;
  int columnSpace;
  int lineSpace;
  int speed;
  int tics;
} AppData;

static AppData app_data;

static XtResource Resources[] = {
  {
    XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData,background),
    XtRString,"XtDefaultBackground"
  },
  {
    XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
    XtOffsetOf(AppData,foreground),
    XtRString,"XtDefaultForeground"
  },
  {
    "message","Message",XtRString,sizeof(String),
    XtOffsetOf(AppData,message),
    XtRString,NULL
  },
  {
    "vfontcap","Vfontcap",XtRString,sizeof(String),
    XtOffsetOf(AppData,vfontcap),
    XtRString,NULL
  },
  {
    "vectorFont","VectorFont",XtRString,sizeof(String),
    XtOffsetOf(AppData,vectorFont),
    XtRString,"min"
  },
  {
    "fontWidth","FontWidth",XtRInt,sizeof(int),
    XtOffsetOf(AppData,fontWidth),
    XtRImmediate,(XtPointer)32
  },
  {
    "fontHeight","FontHeight",XtRInt,sizeof(int),
    XtOffsetOf(AppData,fontHeight),
    XtRImmediate,(XtPointer)32
  },
  {
    "columnSpace","ColumnSpace",XtRInt,sizeof(int),
    XtOffsetOf(AppData,columnSpace),
    XtRImmediate,(XtPointer)0
  },
  {
    "lineSpace","LineSpace",XtRInt,sizeof(int),
    XtOffsetOf(AppData,lineSpace),
    XtRImmediate,(XtPointer)2
  },
  {
    "tics","Tics",XtRInt,sizeof(int),
    XtOffsetOf(AppData,tics),
    XtRImmediate,(XtPointer)50
  },
  {
    "speed","Speed",XtRInt,sizeof(int),
    XtOffsetOf(AppData,speed),
    XtRImmediate,(XtPointer)4
  },
};

static XrmOptionDescRec Options[] = {
  {"-msg","*message",XrmoptionSepArg,(caddr_t)NULL }, 
  {"-vfcap","*vfontcap",XrmoptionSepArg,(caddr_t)NULL },
  {"-vfont","*vectorFont",XrmoptionSepArg,(caddr_t)NULL },
  {"-fw","*fontWidth",XrmoptionSepArg,(caddr_t)NULL },
  {"-fh","*fontHeight",XrmoptionSepArg,(caddr_t)NULL },
  {"-cs","*columnSpace",XrmoptionSepArg,(caddr_t)NULL },
  {"-ls","*lineSpace",XrmoptionSepArg,(caddr_t)NULL },
  {"-tics","*tics",XrmoptionSepArg,(caddr_t)NULL },
  {"-speed","*speed",XrmoptionSepArg,(caddr_t)NULL },
};


/* message window create and control */

typedef struct{
	/* user parameters */

	unsigned char *msg;	/* message(must be EUC-japan) */
	char *font;		/* VFlib font name */
	int size_h;		/* character horizontal size */
	int size_v;		/* character vertical size */
	int space_h;		/* horizontal space beetween characters */
	int space_v;		/* vertical space beetween characters */
	int speed;		/* move speed */

	/* internal parameters */

	int fg;			/* foreground */
	int bg;			/* background */
	int x;			/* position x */
	int y;			/* position y */
	Window window;		/* window */
	Pixmap pixmap;		/* pixmap */
	int width;		/* window width */
	int height;		/* window height */
}msg_t;

static Window Root,Saver;
static msg_t Msg;

static Display *Disp;
static int Scr;
static Colormap Cmap;
static unsigned long Fg,Bg;
static XtAppContext App;
static int Dep;
static int Scr_width,Scr_height;
static GC Default_GC;

static int SS_event, SS_error;

static Saver_state;

static int X_error;


/* error handler */

int err_handler(Display *d,XErrorEvent *e)
{
  X_error = 1;
}


/* create message window */

static int create_msg( msg_t *m )
{
  int column,row;
  int h,v,hs,vs;
  int bufwid,bufsiz;
  unsigned char *buf;
  int space_h,space_v;

  space_h = m->size_h + m->space_h > 0 ? m->space_h : -m->size_h;
  space_v = m->size_v + m->space_v > 0 ? m->space_v : -m->size_v;
  h = m->size_h + space_h;
  v = m->size_v + space_v;

  FONT_get_size(m->msg,&column,&row);
  if( column == 0 || row == 0 ){
    fprintf(stderr,"%s: message has 0 size.\n",Myname);
    return -1;
  }
  hs = h*(column-1)+m->size_h;
  vs = v*(row-1)+m->size_v;

  bufwid = (hs+7)/8;
  bufsiz = bufwid * vs;
  buf = (unsigned char*)malloc(bufsiz);
  if( buf == NULL ){
    fprintf(stderr,"%s: cannot allocate memory.\n",Myname);
    return -1;
  }
  memset(buf,0,bufsiz);
  if( FONT_get_bitmap(m->msg,m->font,buf,bufwid,
  			m->size_h,m->size_v,h,v) < 0 ){
    free(buf);
    return -1;
  }
  FONT_bit_reorder(buf,buf,bufsiz);

  m->window = XCreateSimpleWindow(Disp,Saver,0,0,hs,vs,0,m->bg,m->bg);
  XSelectInput(Disp,m->window,ExposureMask);
  m->pixmap = XCreatePixmapFromBitmapData(Disp,Root,buf,hs,vs,m->fg,m->bg,Dep);
  free(buf);
  XMoveWindow(Disp,m->window,-hs,0);
  XMapWindow(Disp,m->window);
 
  m->width = hs;
  m->height = vs;
  m->x = -m->width;

  return 0;
}


/* move message window */

static void move_msg( msg_t *m )
{
  int range;

  if( m->x < -m->width ){
    m->x = Scr_width;
    range = Scr_height - m->height;
    if( range > 0 )
      m->y = random() % range;
    else
      m->y = -(random() % (-range));
  }
  m->x -= m->speed;
  XMoveWindow(Disp,m->window,m->x,m->y);
}


/* timeout handler */

static void timeout( XtPointer p,XtIntervalId *id )
{
  if( Saver_state ){
    move_msg((msg_t*)p);
    XtAppAddTimeOut(App,app_data.tics,timeout,p);
  }
}


/* start saver */

static void start_saver( msg_t *m )
{
  if( Saver_state )
    return;

  if( create_msg(m) < 0 ){
    FONT_deinit();
    exit(1);
  }

  Saver_state = 1;
  m->x = -m->width;
  move_msg(m);
  XtAppAddTimeOut(App, app_data.tics, timeout,(XtPointer)m);
}


/* stop savcer */

static void stop_saver( msg_t *m )
{
  Saver_state = 0;
  XFreePixmap(Disp,m->pixmap);
}

/* redraw window */

static void redraw( msg_t *m, XEvent *e )
{
  XCopyArea(Disp,m->pixmap,m->window,Default_GC,
  		e->xexpose.x,e->xexpose.y,
  		e->xexpose.width,e->xexpose.height,
  		e->xexpose.x,e->xexpose.y);
}


/* usage */

static void usage( void )
{
  fprintf(stderr,
 	"usage :  %s [-msg message] [-fg color] [-bg color]\n"
 	"\t\t\t[-vfcap vfontcap] [-vfont font]\n"
 	"\t\t\t[-fw font_width] [-fh font_height]\n"
 	"\t\t\t[-cs column_space] [-ls lline_space]\n"
 	"\t\t\t[-speed speed] [-tics tics]\n",
 	Myname);
}

int main( int argc, char **argv )
{
  Widget top;
  XEvent event;
  XScreenSaverNotifyEvent *sevent;
  XScreenSaverInfo *info;
  Pixmap pix_blank;
  GC gc;
  XSetWindowAttributes attr;
  unsigned long mask;
  XColor dummy;
  unsigned char buf[64];
  int (*old_handler)(Display*,XErrorEvent*);
  XID kill_id;
  Atom kill_type;

  Myname = argv[0] ? argv[0] : "xmsgsaver";

  srandom(time(NULL)^getpid());

  top = XtVaAppInitialize(&App,"XMsgSaver",Options,XtNumber(Options),
				&argc,argv,NULL,NULL);
  XtGetApplicationResources(top,&app_data,
  				Resources,XtNumber(Resources),NULL,0);

  if( argc > 1 ){
    usage();
    return 1;
  }


  Disp = XtDisplay(top);
  Root = DefaultRootWindow(Disp);
  Scr = DefaultScreen(Disp);
  Dep = DefaultDepth(Disp,Scr);
  Scr_width = DisplayWidth(Disp,Scr);
  Scr_height = DisplayHeight(Disp,Scr);
  Cmap = DefaultColormap(Disp,Scr);
  Default_GC = DefaultGC(Disp,Scr);

  if( !XScreenSaverQueryExtension(Disp,&SS_event,&SS_error) ){
    fprintf(stderr,"%s: MIT-SCREEN_SAVER is not supported.\n",Myname);
    return 1;
  }
  XScreenSaverSelectInput(Disp,Root,ScreenSaverNotifyMask);

  pix_blank = XCreatePixmap(Disp,Root,1,1,1);
  gc = XCreateGC(Disp,pix_blank,0,NULL);
  XSetForeground(Disp,gc,0);
  XFillRectangle(Disp,pix_blank,gc,0,0,1,1);

  XFreeGC(Disp,gc);

  info = XScreenSaverAllocInfo();
  XScreenSaverQueryInfo(Disp,Root,info);

  if( app_data.message == NULL ){
    gethostname(buf,sizeof(buf));
    Msg.msg = buf;
  }else{
    Msg.msg = app_data.message;
  }

  Msg.size_h = app_data.fontWidth;
  Msg.size_v = app_data.fontHeight;
  Msg.space_h = app_data.columnSpace;
  Msg.space_v = app_data.lineSpace;
  Msg.fg = app_data.foreground;
  Msg.bg = app_data.background;
  Msg.speed = app_data.speed;
  Msg.font = app_data.vectorFont;

  mask = 0;
  attr.colormap = Cmap;
  mask |= CWColormap;
  attr.background_pixel = app_data.background;
  mask |= CWBackPixel;
  attr.cursor = XCreatePixmapCursor(Disp,pix_blank,pix_blank,
					&dummy,&dummy,0,0);
  mask |= CWCursor;


  old_handler = XSetErrorHandler(err_handler);
  X_error = 0;
  XScreenSaverSetAttributes(Disp,Root,0,0,
      				Scr_width,Scr_height,0,
        			CopyFromParent,CopyFromParent,CopyFromParent,
				mask, &attr);
  XSync(Disp,False);
  XSetErrorHandler(old_handler);

  if( X_error ){
    if(XScreenSaverGetRegistered(Disp,Scr,&kill_id, &kill_type)){
      XKillClient(Disp,kill_id);
      XSync(Disp,False);
    }
    XScreenSaverSetAttributes(Disp,Root,0,0,
      				Scr_width,Scr_height,0,
        			CopyFromParent,CopyFromParent,CopyFromParent,
				mask, &attr);
  }

  XScreenSaverRegister (Disp,Scr,(XID)pix_blank,XA_PIXMAP);

  if( FONT_init(app_data.vfontcap) < 0 ){
    return -1;
  }

  XSetErrorHandler(err_handler);

  Saver = info->window;
  if(info->state == ScreenSaverOn){
    if(info->kind != ScreenSaverExternal){
      XResetScreenSaver(Disp);
      XActivateScreenSaver(Disp);
    }
    start_saver(&Msg);
  }

  for(;;){
    XtAppNextEvent(App,&event);
    if(event.type == SS_event){
      sevent = (XScreenSaverNotifyEvent*)&event;
      if(sevent->state == ScreenSaverOn){
        if(sevent->kind != ScreenSaverExternal){
          XResetScreenSaver(Disp);
          XActivateScreenSaver(Disp);
        }else{
          start_saver(&Msg);
        }
      }else if(sevent->state == ScreenSaverOff) {
        stop_saver(&Msg);
      }
    }else if( event.type == Expose && event.xexpose.window == Msg.window ){
      redraw(&Msg,&event);
    }else{
      XtDispatchEvent(&event);
    }
  }	
  /* NOTREACHED */
  return 1;
}
